/*  cluster.cc

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

#include <vector>
#include <algorithm>
#include <iomanip>
#include "fmribmain.h"
#include "newimageall.h"
#include "infer.h"
#include "options.h"

#define _GNU_SOURCE 1
#define POSIX_SOURCE 1

using namespace NEWIMAGE;
using std::vector;
using namespace Utilities;

string title="cluster (Version 1.0)\nCopyright(c) 2000, University of Oxford (Mark Jenkinson)";
string examples="cluster --in <filename> --thresh <value> [options]";

bool verbose = true;
bool help = true;
bool medxcoords = false;
bool mm = false;
bool minv = false;
bool fractional = false;
bool no_table = false;
Option<int> voxvol(string("--volume"), 0,
		   string("number of voxels in the mask"),
		   false, requires_argument);
Option<int> numconnected(string("--connectivity"), 26,
		   string("the connectivity of voxels (default 26)"),
		   false, requires_argument);
Option<float> dLh(string("-d,--dlh"), 4.0,
		  string("smoothness estimate = sqrt(det(Lambda))"),
		  false, requires_argument);
Option<float> thresh(string("-t,--thresh,--zthresh"), 2.3,
		     string("threshold for input volume"),
		     true, requires_argument);
Option<float> pthresh(string("-p,--pthresh"), 0.01,
		      string("p-threshold for clusters"),
		      false, requires_argument);
Option<string> inputname(string("-i,--in,-z,--zstat"), string(""),
			 string("filename of input volume"),
			 true, requires_argument);
Option<string> copename(string("-c,--cope"), string(""),
			string("filename of input cope volume"),
			false, requires_argument);
Option<string> outindex(string("-o,--oindex"), string(""),
			string("filename for output of cluster index (in size order)"),
			false, requires_argument);
Option<string> outthresh(string("--othresh"), string(""),
			 string("filename for output of thresholded image"),
			 false, requires_argument);
Option<string> outsize(string("--osize"), string(""),
		       string("filename for output of size image"),
		       false, requires_argument);
Option<string> outmax(string("--omax"), string(""),
		      string("filename for output of max image"),
		      false, requires_argument);
Option<string> outmean(string("--omean"), string(""),
		       string("filename for output of mean image"),
		       false, requires_argument);
Option<string> transformname(string("-x,--xfm"), string(""),
		       string("filename for transform of input->talairach volume"),
		       false, requires_argument);
Option<string> talvolname(string("--talvol"), string(""),
		       string("filename for talairach volume"),
		       false, requires_argument);



template <class T>
struct triple { T x; T y; T z; };

template <class T, class S>
void copyconvert(const vector<triple<T> >& oldcoords, 
		 vector<triple<S> >& newcoords)
{
  newcoords.erase(newcoords.begin(),newcoords.end());
  newcoords.resize(oldcoords.size());
  for (unsigned int n=0; n<oldcoords.size(); n++) {
    newcoords[n].x = (S) oldcoords[n].x;
    newcoords[n].y = (S) oldcoords[n].y;
    newcoords[n].z = (S) oldcoords[n].z;
  }
}

template <class T, class S>
void vox2medxcoord(vector<triple<T> >& coords, const volume<S>& vol)
{
  int ymax = vol.ysize();
  for (unsigned int n=0; n<coords.size(); n++) {
    coords[n].y = (T) (ymax - 1) - coords[n].y;
  }
}

void vox2mmcoord(vector<triple<float> >& coords, 
				   float vx, float vy, float vz)
{
  for (unsigned int n=0; n<coords.size(); n++) {
    coords[n].x = coords[n].x * vx;
    coords[n].y = coords[n].y * vy;
    coords[n].z = coords[n].z * vz;
  }
}


template <class S>
void vox2mmcoord(vector<triple<float> >& coords, 
				   const volume<S>& vol)
{
  float vx = vol.xdim()*vol.xsign(), vy = vol.ydim()*vol.ysign(), 
    vz = vol.zdim()*vol.zsign();
  vox2mmcoord(coords,vx,vy,vz);
}


//  template <class T, class S>
//  void addoriginoffset(vector<triple<T> >& coords, const volume<S>& vol)
//  {
//    ColumnVector orig;
//    orig = vol.getorigin();
//    if (orig.SumAbsoluteValue() < 0.1)  return;  // zero origin -> no work
//    for (unsigned int n=0; n<coords.size(); n++) {
//      coords[n].x -= (T) (orig(1) - 1.0);   // -1.0 convert from SPM convention
//      coords[n].y -= (T) (orig(2) - 2.0);   // -2.0 convert from SPM convention (NFI)
//      coords[n].z -= (T) (orig(3) - 1.0);   // -1.0 convert from SPM convention
//    }
//  }


//  template <class T, class S>
//  void applyvoxelxfm(vector<triple<T> >& coords, const Matrix& mat, 
//  	      const volume<S>& source, const volume<S>& dest)
//  {
//    Matrix voxmat;
//    voxmat = dest.sampling_mat().i() * mat * source.sampling_mat();
//    ColumnVector vec(4);
//    for (unsigned int n=0; n<coords.size(); n++) {
//      vec << coords[n].x << coords[n].y << coords[n].z << 1.0;
//      vec = voxmat * vec;     // apply voxel xfm
//      coords[n].x = vec(1);
//      coords[n].y = vec(2);
//      coords[n].z = vec(3);
//    }
//  }


template <class T>
void get_stats(const volume<int>& labelim, const volume<T>& origim,
	       vector<int>& size,
	       vector<T>& maxvals, vector<T>& meanvals,
	       vector<triple<int> >& max, vector<triple<float> >& cog,
	       bool minv) 
{
  int labelnum = labelim.max();
  size.resize(labelnum+1,0);
  maxvals.resize(labelnum+1, (T) 0);
  meanvals.resize(labelnum+1,(T) 0);
  triple<int> zero;
  zero.x = 0; zero.y = 0; zero.z = 0;
  triple<float> zerof;
  zerof.x = 0; zerof.y = 0; zerof.z = 0;
  max.resize(labelnum+1,zero);
  cog.resize(labelnum+1,zerof);
  vector<float> sum(labelnum+1,0.0);
  for (int z=labelim.minz(); z<=labelim.maxz(); z++) {
    for (int y=labelim.miny(); y<=labelim.maxy(); y++) {
      for (int x=labelim.minx(); x<=labelim.maxx(); x++) {
	int idx = labelim(x,y,z);
	T oxyz = origim(x,y,z);
	size[idx]++;
	cog[idx].x+=((float) oxyz)*x;
	cog[idx].y+=((float) oxyz)*y;
	cog[idx].z+=((float) oxyz)*z;
	sum[idx]+=(float) oxyz;
	if ( (size[idx]==1) || 
	     ( (oxyz>maxvals[idx]) && (!minv) ) || 
	     ( (oxyz<maxvals[idx]) && (minv) ) ) 
	  {
	    maxvals[idx] = oxyz;
	    max[idx].x = x;
	    max[idx].y = y;
	    max[idx].z = z;
	  }
      }
    }
  }
  for (int n=0; n<=labelnum; n++) {
    if (size[n]>0.0) {
      meanvals[n] = (T) (sum[n]/((float) size[n]));
    }
    if (sum[n]>0.0) {
      cog[n].x /= sum[n];
      cog[n].y /= sum[n];
      cog[n].z /= sum[n];
    }
  }
}


vector<int> get_sortindex(const vector<int>& vals)
{
  // return the mapping of old indices to new indices in the
  //   new *ascending* sort of vals
  int length=vals.size();
  vector<pair<int, int> > sortlist(length);
  for (int n=0; n<length; n++) {
    sortlist[n] = pair<int, int>(vals[n],n);
  }
  sort(sortlist.begin(),sortlist.end());  // O(N.log(N))
  vector<int> idx(length);
  for (int n=0; n<length; n++) {
    idx[n] = sortlist[n].second;
  }
  return idx;
}


void get_sizeorder(const vector<int>& size, vector<int>& sizeorder) 
{
  vector<int> sizecopy(size), idx;
  idx = get_sortindex(sizecopy);

  // second part of pair is now the prior-index of the sorted values
  int length = size.size();
  sizeorder.resize(length,0);
  for (int n=0; n<length; n++) {
    sizeorder[idx[n]] = n;    // maps old index to new
  }
}


template <class T, class S>
void relabel_image(const volume<int>& labelim, volume<T>& relabelim,
		   const vector<S>& newlabels)
{
  copyconvert(labelim,relabelim);
  for (int z=relabelim.minz(); z<=relabelim.maxz(); z++) {
    for (int y=relabelim.miny(); y<=relabelim.maxy(); y++) {
      for (int x=relabelim.minx(); x<=relabelim.maxx(); x++) {
	relabelim(x,y,z) = (T) newlabels[labelim(x,y,z)];
      }
    }
  }
}


template <class T, class S>
void print_results(const vector<int>& idx, 
		   const vector<int>& size,
		   const vector<int>& pthreshindex,
		   const vector<float>& pvals,
		   const vector<T>& maxvals,
		   const vector<triple<S> >& maxpos,
		   const vector<triple<float> >& cog,
		   const vector<T>& copemaxval,
		   const vector<triple<S> >& copemaxpos,
		   const vector<T>& copemean,
		   const volume<T>& zvol, const volume<T>& cope)
{
  int length=size.size();

  vector<triple<float> > fmaxpos, fcopemaxpos, fcog;
  copyconvert(maxpos,fmaxpos);
  copyconvert(cog,fcog);
  copyconvert(copemaxpos,fcopemaxpos);
  volume<T> talvol;
  if ( !transformname.unset() && !talvolname.unset() ) {
    read_volume(talvol,talvolname.value());
//      Matrix trans;
//      read_matrix(trans,transformname.value(),zvol);
    if (verbose) { 
      cout << "Transformation Matrix filename = "<<transformname.value()<<endl;
//        cout << trans.Nrows() << " " << trans.Ncols() << endl; 
//        cout << "Transformation Matrix = " << endl;
//        cout << trans << endl;
    }
//      applyvoxelxfm(fmaxpos,trans,zvol,talvol);
//      applyvoxelxfm(fcog,trans,zvol,talvol);
//      if (!copename.unset()) applyvoxelxfm(fcopemaxpos,trans,zvol,talvol);
  }
  if ( medxcoords && (!mm) ) {
    vox2medxcoord(fmaxpos,zvol);
    vox2medxcoord(fcog,zvol);
    if (!copename.unset()) vox2medxcoord(fcopemaxpos,cope);
  }
  if (mm) {
    const volume<T> *refvol = &zvol;
    if ( !transformname.unset() && !talvolname.unset() ) refvol = &talvol;
    if (verbose) { 
      cout << "Using origin : " << (refvol->getorigin()).t() << endl; 
    }
//      addoriginoffset(fmaxpos,*refvol);
//      addoriginoffset(fcog,*refvol);
//      if (!copename.unset()) addoriginoffset(fcopemaxpos,*refvol);
    vox2mmcoord(fmaxpos,*refvol);
    vox2mmcoord(fcog,*refvol);
    if (!copename.unset()) vox2mmcoord(fcopemaxpos,*refvol);  // used cope before
  }

  string units=" (vox)";
  if (!mm) units=" (mm)";
  string tablehead;
  tablehead = "Cluster Index\tVoxels";
  if (!pthresh.unset()) tablehead += "\tP\t-log10(P)";
  tablehead += "\tMax Z\tX" + units + "\tY" + units + "\tZ" + units
    + "\tCOG X" + units + "\tCOG Y" + units + "\tCOG Z" + units;
  if (!copename.unset()) {
    tablehead+= "\tMax COPE\tX" + units + "\tY" + units + "\tZ" + units 
                 + "\tMean COPE";
  }

  cout << tablehead << endl;

  for (int n=length-1; n>=1; n--) {
    int index=idx[n];
    int k = size[index];
    float p = pvals[index];
    if (pthreshindex[index]>0) {
      float mlog10p;
      if (p>1e-38) mlog10p=-log10(p);
      else mlog10p=42;
      cout << setprecision(3) << pthreshindex[index] << "\t" << k << "\t"; 
      if (!pthresh.unset()) { cout << p << "\t" << mlog10p << "\t"; }
      cout << maxvals[index] << "\t" 
	   << fmaxpos[index].x << "\t" << fmaxpos[index].y << "\t" 
	   << fmaxpos[index].z << "\t"
	   << fcog[index].x << "\t" << fcog[index].y << "\t" 
	   << fcog[index].z;
      if (!copename.unset()) { 
	cout   << "\t" << copemaxval[index] << "\t"
	       << fcopemaxpos[index].x << "\t" << fcopemaxpos[index].y << "\t" 
	       << fcopemaxpos[index].z << "\t" << copemean[index];
	}
      cout << endl;
    }
  }
}



template <class T>
int fmrib_main(int argc, char *argv[])
{
  volume<int> labelim;
  vector<int> size;
  vector<triple<int> > maxpos;
  vector<triple<float> > cog;
  vector<T> maxvals, meanvals;
  float th = thresh.value();

  // read in the volume
  volume<T> zvol, mask, cope;
  read_volume(zvol,inputname.value());
  if (verbose)  print_volume_info(zvol,"Zvol");
  
  if ( fractional ) {
    float frac = th;
    th = frac*(zvol.robustmax() - zvol.robustmin()) + zvol.robustmin();
  }
  mask = zvol;
  mask.threshold((T) th);
  if (minv) { mask = ((T) 1) - mask; }
  if (verbose)  print_volume_info(mask,"Mask");
  
  // Find the connected components
  labelim = connected_components(mask,numconnected.value());
  if (verbose)  print_volume_info(labelim,"Labelim");
  
  // process according to the output format
  get_stats(labelim,zvol,size,maxvals,meanvals,maxpos,cog,minv);
  if (verbose) {cout<<"Number of labels = "<<size.size()<<endl;}

  // process the cope image if entered
  vector<int> copesize;
  vector<triple<int> > copemaxpos;
  vector<triple<float> > copecog;
  vector<T> copemaxval, copemean;
  if (!copename.unset()) {
    read_volume(cope,copename.value());
    get_stats(labelim,cope,copesize,copemaxval,copemean,copemaxpos,copecog,
	      minv);
  }

  // re-threshold for p
  int length = size.size();
  size[0] = 0;  // force background to be ordered last
  vector<int> pthreshsize;
  vector<float> pvals(length);
  pthreshsize = size;
  int nozeroclust=0;
  if (!pthresh.unset()) {
    if (verbose) {cout<<"Re-thresholding with p-value"<<endl;}
    Infer infer(dLh.value(), th, voxvol.value());
    for (int n=1; n<length; n++) {
      int k = size[n];
      pvals[n] = infer(k);
      if (pvals[n]>pthresh.value()) {
	pthreshsize[n] = 0;
	nozeroclust++;
      }
    }
  }
  if (verbose) {cout<<"Number of sub-p clusters = "<<nozeroclust<<endl;}

  // get sorted index (will revert to cluster size if no pthresh)
  vector<int> idx;
  idx = get_sortindex(pthreshsize);
  if (verbose) {cout<<pthreshsize.size()<<" labels in sortedidx"<<endl;}

  vector<int> threshidx(length);
  for (int n=length-1; n>=1; n--) {
    int index=idx[n];
    if (pthreshsize[index]>0) {
      threshidx[index] = n - nozeroclust;
    }
  }

  // print table
  if (!no_table) {
    print_results(idx, size, threshidx, pvals, maxvals, maxpos, cog,
		  copemaxval, copemaxpos, copemean, zvol, cope);
  }

  // save relevant volumes
  if (!outindex.unset()) {
    volume<int> relabeledim;
    relabel_image(labelim,relabeledim,threshidx);
    save_volume(relabeledim,outindex.value());
  }
  if (!outsize.unset()) {
    volume<int> relabeledim;
    relabel_image(labelim,relabeledim,pthreshsize);
    save_volume(relabeledim,outsize.value());
  }
  if (!outmax.unset()) {
    volume<int> relabeledim;
    relabel_image(labelim,relabeledim,maxvals);
    save_volume(relabeledim,outmax.value());
  }
  if (!outmean.unset()) {
    volume<int> relabeledim;
    relabel_image(labelim,relabeledim,meanvals);
    save_volume(relabeledim,outmean.value());
  }
  if (!outthresh.unset()) {
    // Threshold the input volume st it is 0 for all non-clusters
    //   and maintains the same values otherwise
    volume<T> lcopy;
    relabel_image(labelim,lcopy,threshidx);
    lcopy.threshold(1);
    save_volume(zvol * lcopy,outthresh.value());
  }

  return 0;
}



int main(int argc,char *argv[])
{

  Tracer tr("main");

  OptionParser options(title, examples);

  try {
    options.add(inputname);
    options.add(thresh);
    options.add(outindex);
    options.add(outthresh);
    options.add(outsize);
    options.add(outmax);
    options.add(outmean);
    options.add(pthresh);
    options.add(copename);
    options.add(voxvol);
    options.add(dLh);
//      options.add(fractional);
    options.add(numconnected);
//      options.add(medxcoords);
//      options.add(mm);
//      options.add(minv);
//      options.add(no_table);
    options.add(transformname);
    options.add(talvolname);
//      options.add(verbose);
//      options.add(help);
    
    options.parse_command_line(argc, argv);

    if ( (help) || (!options.check_compulsory_arguments(true)) )
      {
	options.usage();
	exit(EXIT_FAILURE);
      }
    
    if ( (!pthresh.unset()) && (dLh.unset() || voxvol.unset()) ) 
      {
	options.usage();
	cerr << endl 
	     << "Both --dlh and --volume MUST be set if --pthresh is used." 
	     << endl;
	exit(EXIT_FAILURE);
      }
    
    if ( ( !transformname.unset() && talvolname.unset() ) ||
	 ( transformname.unset() && (!talvolname.unset()) ) ) 
      {
	options.usage();
	cerr << endl 
	     << "Both --xfm and --talvol MUST be set together." 
	     << endl;
	exit(EXIT_FAILURE);
      }
  }  catch(X_OptionError& e) {
    options.usage();
    cerr << endl << e.what() << endl;
    exit(EXIT_FAILURE);
  } catch(std::exception &e) {
    cerr << e.what() << endl;
  } 

  return call_fmrib_main(dtype(inputname.value()),argc,argv);

}

