/*  f2z.cc

    Mark Woolrich & Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

#include <cmath>
#include "f2z.h"
#include "Log.h"
#include "glm.h"
#include <stdexcept>
extern "C"
{
#include "libprob.h"
};

using namespace UTILS;
using namespace NEWMAT;
using namespace SIGPROC;
//#define MAXFLOAT 1e+36

namespace UTILS {

  F2z* F2z::f2z = NULL;
 
  float F2z::largef2logp(float f, int d1, int d2)
  {
    // no of iterations:
    int N = 20;

    if (f<=0.0) {
      cerr << "f cannot be zero or negative!" << endl;
      return 0.0;
    }
  
    if (d1<=0 || d2<=0) { 
      cerr << "DOFs cannot be zero or negative!" << endl;
      return 0.0; 
    }
 
    double alpha=d1/(double)d2;
    double m=(d1+d2)/2.0;
    double n=(1-d1/2.0);
    double loggam = (d1/2.0)*log(d1/(double)d2)-logbeta(d2/2.0,d1/2.0);

    //iter=f^(-n)/(alpha*(n+m-1)) + n*f^(-(n+1))/(alpha^2*(n+m-1)*(n+m)) + n*(n+1)*f^(-(n+2))/(alpha^3*(n+m-1)*(n+m)*(n+m+1));

    double top = 1.0;
    double bot = n+m-1;
    double iter = 0.0;

    for(int i = 1; i <= N; i++)
      {
	iter = iter + top*pow(f,(-(n+i-1)))/(pow(alpha,i)*bot);
	top = top*(n-1+i)*(-1);
	bot = bot*(n+m-1+i);
      }

    float logp = loggam-(m-1)*log(1+alpha*f)+log(iter);

    //      cerr << "iter = " << iter << endl;
    //      cerr << "loggam = " << loggam << endl;
    //      cerr << "n = " << n << endl;
    //      cerr << "m = " << m << endl;
    //      cerr << "logp = " << logp << endl;
    //      cerr << "logbeta(d2/2.0,d1/2.0)=" << logbeta(d2/2.0,d1/2.0) << endl;
    //      cerr << "lgam(5.0)=" << lgam(5.0) << endl;

    return logp;
  }

  bool F2z::islargef(float f, int d1, int d2, float &logp) {
    
    logp=largef2logp(f,d1,d2);
    
    return issmalllogp(logp);
  }

  bool F2z::issmalllogp(float logp) {
    return (logp < -14.5);
  }

  float F2z::convert(float f, int d1, int d2) 
  {
    float z = 0.0, logp=0.0;

    if(!islargef(f,d1,d2,logp)) {

      double p = fdtr(d1, d2, f);

      z = ndtri(p);
    }
      else {

	z = logp2largez(logp);
      }

      return z;
    }

  void F2z::ComputeZStats(const ColumnVector& p_fs, int p_dof1, int p_dof2, ColumnVector& p_zs)
    {
      Tracer ts("F2z::ComputeStats");

      int numTS = p_fs.Nrows();

      p_zs.ReSize(numTS);
      F2z& f2z = F2z::getInstance();

      for(int i = 1; i <= numTS; i++)
	{
	  if (p_fs(i) > 0.0)
	    {
	      p_zs(i) = f2z.convert(p_fs(i),p_dof1,p_dof2);  
	    }
	  else
	    {
	      p_zs(i) = 0.0;
	    }     
	}
    }

}






























