/*  sigproc.cc

    Mark Woolrich, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

#include "ols.h"
#include "sigproc.h"
#include "miscmaths.h"
#include "glm.h"

#ifndef NO_NAMESPACE
namespace SIGPROC {
  using namespace MISCMATHS;
#endif

  void Out(string p_fname, const Matrix& p_mat)
    {
      ofstream out;
      out.open(p_fname.c_str(), ios::out);
      out << "/Matrix" << endl;
      out << p_mat;
      out.close();
    }

  void Out(string p_fname, const ColumnVector& p_mat)
    {
      ofstream out;
      out.open(p_fname.c_str(), ios::out);
      out << p_mat;
      out.close();
    }

  void Out(string p_fname, const RowVector& p_mat)
    {
      ofstream out;
      out.open(p_fname.c_str(), ios::out);
      out << p_mat;
      out.close();
    }

  void In(string p_fname, ColumnVector& p_mat)
    {
      int size = p_mat.Nrows();

      ifstream in;
      in.open(p_fname.c_str(), ios::in);
      
      if(!in)
	{
	  string err("Unable to open ");
	  err =  err +  p_fname;
	  throw Exception(err.c_str());
	}

      for(int i = 1; i <= size; i++)
	{
	  in >> p_mat(i);
	}

      in.close();
    }

  void In(string p_fname, Matrix& p_mat)
    {
      ifstream in;
      in.open(p_fname.c_str(), ios::in);
      
      if(!in)
	{
	  cerr << "Unable to open " << p_fname << endl;
	  throw;
	}

      int numWaves = 0;
      int numPoints = 0;
       
      string str;

      while(true)
	{
	  in >> str;
	  if(str == "/Matrix")
	    break;
	  else if(str == "/NumWaves")
	    {
	      in >> numWaves;
	    }
	  else if(str == "/NumPoints" || str == "/NumContrasts")
	    {
	      in >> numPoints;
	    }
	}

      if(numWaves != 0)
	{
	  p_mat.ReSize(numPoints, numWaves);
	}
      else
	{
	  numWaves = p_mat.Ncols();
	  numPoints = p_mat.Nrows();
	}
      
      for(int i = 1; i <= numPoints; i++)
	{
	  for(int j = 1; j <= numWaves; j++)    
	    {
	      in >> p_mat(i,j);
	    }
	  }

      in.close();
    }

  int EstablishZeroPadding(int sizeTS) {
    
    Tracer tr("EstablishZeroPadding");
    
    return (int)pow(2,ceil(log(sizeTS)/log(2)));
  }

  void BandPass(Matrix& p_ts, int lowcut, int highcut)
  {
    Tracer tr("BandPass");

      int sizeTS = p_ts.Nrows();
      int numTS = p_ts.Ncols();
      
      int zeropad = EstablishZeroPadding(sizeTS);
      ColumnVector x(zeropad);
      x = 0;
      ColumnVector fft_real;
      ColumnVector fft_im;
      ColumnVector dummy(zeropad);
      ColumnVector dummy2;
      dummy = 0;
      ColumnVector realifft(zeropad);

      // convert cutofs to zeropad cutoffs:
      lowcut = (int)(lowcut*zeropad/(float)sizeTS);
      highcut = (int)(highcut*zeropad/(float)sizeTS);

      for(int i = 1; i <= numTS; i++)
	{
	  x.Rows(1,sizeTS) = p_ts.Column(i);

	  FFT(x, dummy, fft_real, fft_im);

	  // Perform cutoff here:
	  // everything greater than highcut and lower than lowcut is removed
	  if(lowcut > 0 && lowcut < sizeTS){
	    fft_real.Rows(1, lowcut) = 0;
	    fft_im.Rows(1, lowcut) = 0;
	    fft_real.Rows(zeropad - lowcut, zeropad) = 0;
	    fft_im.Rows(zeropad - lowcut, zeropad) = 0;
	  }
	  if(highcut < sizeTS && highcut > 0){
	    fft_real.Rows(highcut, zeropad - highcut) = 0;
	    fft_im.Rows(highcut, zeropad - highcut) = 0;
	  }
	  //////////////////////

	  FFTI(fft_real, fft_im, realifft, dummy2);
	  p_ts.Column(i) = realifft.Rows(1,sizeTS);
	} 
  }

  void AutoCorr(const Matrix& p_ts, Matrix& p_ret, Matrix& fts, int p_zeropad)
    {
      Tracer tr("AutoCorr");

      int sizeTS = p_ts.Nrows();
      int numTS = p_ts.Ncols();

      ColumnVector x(p_zeropad);
      x = 0;
      ColumnVector fft_real;
      ColumnVector fft_im;
      ColumnVector dummy(p_zeropad);
      ColumnVector dummy2;
      dummy = 0;
      ColumnVector realifft(p_zeropad);
      p_ret.ReSize(sizeTS,numTS);
      fts.ReSize(p_zeropad,numTS);

      for(int i = 1; i <= numTS; i++)
	{
	  x.Rows(1,sizeTS) = p_ts.Column(i);
	  FFT(x, dummy, fft_real, fft_im);

	  for(int j = 1; j <= p_zeropad; j++)
	    {
	      // (x+iy)(x-iy) = x^2 + y^2
	      fft_real(j) = fft_real(j)*fft_real(j) + fft_im(j)*fft_im(j);
	      fft_im(j) = 0;
	      fts(j,i) = fft_real(j)*fft_real(j) + fft_im(j)*fft_im(j);      
	    }

	  // normalise fts:
	  fts.Column(i) = fts.Column(i)/fts.Column(i).Sum();
	  
	  FFTI(fft_real, fft_im, realifft, dummy2);

	  float varx = var(ColumnVector(x.Rows(1,sizeTS)));
	  p_ret.Column(i) = realifft.Rows(1,sizeTS);
 
	  for(int j = 1; j <= sizeTS-1; j++)
	    {
	       // Correction to make autocorr unbiased and normalised
	       p_ret(j,i) = p_ret(j,i)/((sizeTS-j)*varx);
	    }  
	}

    }

  int Pacf(const ColumnVector& x, int minorder, int maxorder, ColumnVector& betas)
    { 
      Tracer ts("pacf");
      int order = -1;
      int sizeTS = x.Nrows();

      // Set c
      Matrix c(1,1);
      c(1,1) = 1;
      
      Glm glm;

      for(int i = minorder+1; i <= maxorder+1; i++)
	{
	  ColumnVector y = x.Rows(i+1,sizeTS);

	  // Setup design matrix
	  Matrix X(sizeTS-i, i);
	  X = 0;
	  
	  for(int j = 1; j <= i; j++)
	    {
	      X.Column(j) = x.Rows(i+1-j,sizeTS-j).AsColumn();	    
	    }
	  
	  glm.setData(y, X, c);
	  
	  glm.ComputeResids();
	  betas = glm.Getb();
	  
	  if((abs(betas(i)) < (1/sizeTS) + (2/pow(sizeTS,0.5)) && order == -1) 
	     || i == maxorder+1)
          {      
	    order = i-1;
	    break;
	  }
	}
      return order; 
    }

  void Detrend(Matrix& p_ts, bool p_justremovemean)
    {
      Tracer trace("SIGPROC::Detrend");

      cerr << "Detrending...";
      int sizeTS = p_ts.Nrows();
     
      // Set c
      Matrix c(1,1);
      c(1,1) = 1;

      // p_ts = b*a + p_ret OLS regression
      Matrix a(sizeTS, 3);

      if(p_justremovemean) a.ReSize(sizeTS,1);

      // Create a
      for(int i = 1; i <= sizeTS; i++)
	{
	  a(i,1) = 1;
	  if(!p_justremovemean)
	    {
	      a(i,2) = (float)i/sizeTS;
	      a(i,3) = pow((float)i/sizeTS,2);
	    }
	}
          
      // Do OLS for all TS:
      Ols ols(p_ts, a, c);

      p_ts = ols.ComputeResids();
      
      cerr << " Completed" << endl;
    }

#ifndef NO_NAMESPACE
}
#endif






