/*  miscmaths.h

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

// Miscellaneous maths functions


#if !defined(__miscmaths_h)
#define __miscmaths_h

#include <cmath>
#include <complex>
#include <vector>
#include <iostream>
#include <assert.h>
#include <fstream>
#include <strstream>
#include <string>
#include "newmatap.h"

//#pragma interface

using namespace NEWMAT;

namespace MISCMATHS {


#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif


  typedef std::complex<double> complex;

  // IO/string stuff

  template <class T>
  string num2str(T n, int width=-1);

  bool isnum(const string& str);
  Matrix read_ascii_matrix(const string& filename, int nrows, int ncols);
  Matrix read_ascii_matrix(const string& filename);
  int write_ascii_matrix(const Matrix& mat, const string& filename, 
			 int precision=-1);
  // more basic IO calls
  string skip_alpha(ifstream& fs);
  Matrix read_ascii_matrix(ifstream& fs, int nrows, int ncols);
  Matrix read_ascii_matrix(ifstream& fs);
  int write_ascii_matrix(const Matrix& mat, ofstream& fs, int precision=-1);


  // General maths

  int round(int x); 
  int round(float x);
  int round(double x);

  int periodicclamp(int x, int x1, int x2);

  template<class S, class T> 
   inline T Min(const S &a, const T &b) { if (a<b) return (T) a; else return b; }

  template<class S, class T>
   inline T Max(const S &a, const T &b) { if (a>b) return (T) a; else return b; }

  template<class T>
   inline T Sqr(const T& x) { return x*x; } 

  ColumnVector cross(const ColumnVector& a, const ColumnVector& b);
  ColumnVector cross(const Real *a, const Real *b);

  inline float dot(const ColumnVector& a, const ColumnVector& b)
    { return Sum(SP(a,b)); }

  float norm2(const ColumnVector& x);

  int Identity(Matrix& m);
  Matrix Identity(int num);

  int diag(Matrix& m, const float diagvals[]);
  int diag(Matrix& m, const ColumnVector& diagvals);
  Matrix diag(const Matrix& mat);
  Matrix pinv(const Matrix& mat);
  Matrix sqrtaff(const Matrix& mat);

  void reshape(Matrix& r, const Matrix& m, int nrows, int ncols);

  int construct_rotmat_euler(const ColumnVector& params, int n, Matrix& aff);
  int construct_rotmat_euler(const ColumnVector& params, int n, Matrix& aff,
		   const ColumnVector& centre);
  int construct_rotmat_quat(const ColumnVector& params, int n, Matrix& aff);
  int construct_rotmat_quat(const ColumnVector& params, int n, Matrix& aff,
		   const ColumnVector& centre);
  int make_rot(const ColumnVector& angl, const ColumnVector& centre, 
	       Matrix& rot);

  int getrotaxis(ColumnVector& axis, const Matrix& rotmat);
  int rotmat2euler(ColumnVector& angles, const Matrix& rotmat);
  int rotmat2quat(ColumnVector& quaternion, const Matrix& rotmat);
  int decompose_aff(ColumnVector& params, const Matrix& affmat,
		    int (*rotmat2params)(ColumnVector& , const Matrix& ));
  int decompose_aff(ColumnVector& params, const Matrix& affmat,
		    const ColumnVector& centre,
		    int (*rotmat2params)(ColumnVector& , const Matrix& ));
  int compose_aff(const ColumnVector& params, int n, const ColumnVector& centre,
		  Matrix& aff, 
		  int (*params2rotmat)(const ColumnVector& , int , Matrix& ,
				       const ColumnVector& ) );
  float rms_deviation(const Matrix& affmat1, const Matrix& affmat2, 
		      const ColumnVector& centre, const float rmax); 
  float rms_deviation(const Matrix& affmat1, const Matrix& affmat2, 
		      const float rmax=80.0); 

  // Added by MWW
  float var(const ColumnVector& x);
  float mean(const ColumnVector& x);
  float median(const ColumnVector& x);


  // Added by CFB
  Matrix zeros(const int dim1, const int dim2 = -1);
  Matrix ones(const int dim1, const int dim2 = -1);
  Matrix abs(const Matrix& mat);
  Matrix sqrt(const Matrix& mat);
  Matrix log(const Matrix& mat);
  Matrix exp(const Matrix& mat);
  Matrix tanh(const Matrix& mat);
  Matrix pow(const Matrix& mat, const double exp);
  Matrix sum(const Matrix& mat, const int dim);
  Matrix mean(const Matrix& mat, const int dim = 1);
  Matrix var(const Matrix& mat, const int dim);
  Matrix remmean(const Matrix& mat, const int dim = 1);
  Matrix stdev(const Matrix& mat, const int dim = 1);
  SymmetricMatrix cov(const Matrix& mat, const int norm = 0);
  SymmetricMatrix corrcoef(const Matrix& mat, const int norm = 0);
  void symm_orth(Matrix &Mat);
  Matrix drandm(const int dim1, const int dim2 = -1);

  
  // A complex Matrix class (based on NEWMAT)
  class complexmat {
  private:
     Matrix rmat;
     Matrix imat;
     void copy(const complexmat& source) 
        { rmat = source.rmat; imat = source.imat; }
  public:
     complexmat() { } ;
     complexmat(const Matrix& realmat, const Matrix & imagmat)
        { rmat = realmat; imat = imagmat; }
     complexmat(unsigned int x, unsigned int y)
        { rmat.ReSize(x,y);  imat.ReSize(x,y); }
     complexmat(const complexmat& source) 
        { this->copy(source); }
     ~complexmat() { } ;
     complexmat& operator=(const complexmat& source)
        { this->copy(source);  return *this; }
     complexmat& operator=(const Matrix& source)
        { rmat = source; imat = source*0; return *this; }
     complexmat& operator=(const Real& val)
        { rmat = val; imat = rmat*0; return *this; }
     complexmat& operator=(const complex& val)
        { rmat = std::real(val); imat = std::imag(val); return *this; }

     void ReSize(unsigned int x, unsigned int y) 
        { rmat.ReSize(x,y);  imat.ReSize(x,y); }
     int Nrows() const { return rmat.Nrows(); }
     int Ncols() const { return rmat.Ncols(); }
     complexmat SubMatrix(int fr,int lr, int fc, int lc) const {
     return complexmat(rmat.SubMatrix(fr,lr,fc,lc),imat.SubMatrix(fr,lr,fc,lc));
     }

     const complex operator()(unsigned int x, unsigned int y) const
         { complex res(rmat(x,y),imat(x,y)); return res; }

     Matrix real() const { return rmat; }
     Matrix imag() const { return imat; }
     const Real real(unsigned int x, unsigned int y) const 
        { return rmat(x,y); }
     const Real imag(unsigned int x, unsigned int y) const 
        { return imat(x,y); }
     Real& real(unsigned int x, unsigned int y)
        { return rmat(x,y); }
     Real& imag(unsigned int x, unsigned int y)
        { return imat(x,y); }
     const complex& set(unsigned int x, unsigned int y, const complex& v) 
        { rmat(x,y)=std::real(v); imat(x,y)=std::imag(v); return v; }
     void setreal(const Matrix& mat) { rmat = mat; }  // should check size
     void setimag(const Matrix& mat) { imat = mat; }  // should check size
     void set(const Matrix& realmat, const Matrix & imagmat) 
        { rmat = realmat; imat = imagmat; }  // should check size

     complexmat t() const { return complexmat(rmat.t(),imat.t()); }
     complexmat conj() const { return complexmat(rmat,-imat); }
     complexmat i() const 
        { Matrix c, d;  
	  c = (rmat + imat*rmat.i()*imat).i();  d = -c*imat*rmat.i(); 
          complexmat mnew(c,d);  return mnew; }
     complexmat Reverse() const 
        { return complexmat(rmat.Reverse(),imat.Reverse()); }

     Matrix Absolute() const 
         { Matrix tmp;  tmp = SP(rmat,rmat) + SP(imat,imat); return tmp; }
     const Real MaximumAbsoluteValue() const
         { Matrix tmp;  tmp = this->Absolute(); 
	   return tmp.MaximumAbsoluteValue(); } 
     const Real MinimumAbsoluteValue() const
         { Matrix tmp;  tmp = this->Absolute(); 
	   return tmp.MinimumAbsoluteValue(); } 
     complex Trace() const { return complex(rmat.Trace(),imat.Trace()); }
     complex AsScalar() const { return this->Trace(); }
     complex Sum() const
          { return complex(rmat.Sum(),imat.Sum()); }
     Real SumAbsoluteValue() const { return (this->Absolute()).Sum(); }
     bool IsZero() const 
         { return ((rmat.IsZero()) && (imat.IsZero())); }

     bool operator==(const complexmat& test) const
         { return ((rmat == test.rmat) && (imat == test.imat)); }
     bool operator!=(const complexmat& test) const { return !(*this == test); }

     complexmat operator-() { complexmat res(*this); res*=-1.0; return res; }
     complexmat& operator+=(const Real& v) { rmat += v; return *this; }
     complexmat& operator-=(const Real& v) { rmat -= v; return *this; }
     complexmat& operator*=(const Real& v) 
        { rmat *= v; imat *=v; return *this; }
     complexmat& operator/=(const Real& v) 
        { rmat /= v; imat /=v; return *this; }
     complexmat& operator+=(const complex& v) 
        { rmat += std::real(v); imat += std::imag(v); return *this; }
     complexmat& operator-=(const complex& v) 
        { rmat -= std::real(v); imat -= std::imag(v); return *this; }
     complexmat& operator*=(const complex& v) 
        { Matrix oldr(rmat), oldi(imat); 
	  rmat = std::real(v)*oldr - std::imag(v)*oldi; 
	  imat = std::imag(v)*oldr + std::real(v)*oldi; 
	  return *this; }
     complexmat& operator/=(const complex& v) 
        { Matrix oldr(rmat), oldi(imat); 
	  rmat = std::real(v)*oldr + std::imag(v)*oldi; 
	  imat = std::real(v)*oldi - std::imag(v)*oldr;
	  rmat /= std::norm(v);
	  imat /= std::norm(v);
	  return *this; }
     complexmat operator+(const Real& v) const
        { complexmat mnew(rmat, imat);  mnew += v;  return mnew; }
     complexmat operator-(const Real& v) const 
        { complexmat mnew(rmat, imat);  mnew -= v;  return mnew; }
     complexmat operator*(const Real& v) const 
        { complexmat mnew(rmat, imat);  mnew *= v;  return mnew; }
     complexmat operator/(const Real& v) const 
        { complexmat mnew(rmat, imat);  mnew /= v;  return mnew; }
     complexmat operator+(const complex& v) const 
        { complexmat mnew(rmat, imat);  mnew += v;  return mnew; }
     complexmat operator-(const complex& v) const 
        { complexmat mnew(rmat, imat);  mnew -= v;  return mnew; }
     complexmat operator*(const complex& v) const 
        { complexmat mnew(rmat, imat);  mnew *= v;  return mnew; }
     complexmat operator/(const complex& v) const 
        { complexmat mnew(rmat, imat);  mnew /= v;  return mnew; }

     friend complexmat operator+(const Real& v, const complexmat& mat);
     friend complexmat operator-(const Real& v, const complexmat& mat); 
     friend complexmat operator*(const Real& v, const complexmat& mat); 
     friend complexmat operator+(const complex& v, const complexmat& mat); 
     friend complexmat operator-(const complex& v, const complexmat& mat); 
     friend complexmat operator*(const complex& v, const complexmat& mat); 

     complexmat& operator+=(const complexmat& v)
        { rmat += v.rmat ; imat += v.imat; return *this; }
     complexmat& operator-=(const complexmat& v)
        { rmat -= v.rmat ; imat -= v.imat; return *this; }
     complexmat& operator*=(const complexmat& v)
        { Matrix oldr(rmat), oldi(imat);  
	  rmat = oldr * v.rmat - oldi * v.imat; 
	  imat = oldr * v.imat + oldi * v.rmat; 
	  return *this; }
     complexmat operator+(const complexmat& v) const
        { complexmat mnew(rmat, imat);  mnew += v;  return mnew; }
     complexmat operator-(const complexmat& v) const
        { complexmat mnew(rmat, imat);  mnew -= v;  return mnew; }
     complexmat operator*(const complexmat& v) const
        { complexmat mnew(rmat, imat);  mnew *= v;  return mnew; }

   };

   // helper functions for the Matrix class
   Matrix real(const complexmat& mat);
   Matrix imag(const complexmat& mat);
   complexmat conj(const complexmat& mat);
   Matrix Absolute(const complexmat& mat);
   complex Trace(const complexmat& mat);
   complexmat SP(const complexmat& mat1, const complexmat& mat2);
   complexmat complexify(const Matrix& mat);
   complex AsScalar(const complexmat& mat);
   complex Sum(const complexmat& mat);
   Real SumAbsoluteValue(const complexmat& mat);
   bool IsZero(const complexmat& mat);
   complexmat operator-(const Real& v, const complexmat& mat); 
   complexmat operator*(const Real& v, const complexmat& mat); 
   complexmat operator+(const complex& v, const complexmat& mat); 
   complexmat operator-(const complex& v, const complexmat& mat); 
   complexmat operator*(const complex& v, const complexmat& mat);
   void FFT(const complexmat& X, complexmat& F);
   void FFTI(const complexmat& F, complexmat& X);
   void print(const complexmat& v);


   // stuff that uses complexmat

   typedef std::vector<complexmat> veccomplexmat;

   void matfuncore(const complexmat& m, complexmat& lambda, 
		   ColumnVector& mult,
		   veccomplexmat& beta, veccomplexmat& alpha,
		   veccomplexmat& gamma, veccomplexmat& delta);
   complexmat expm(const complexmat& m);
   complexmat logm(const complexmat& m);
   Matrix expm(const Matrix& m);
   complexmat logm(const Matrix& m);
 
  
  ///////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////
 
  // TEMPLATE DEFINITIONS //
  
  template <class T>
  string num2str(T n, int width)
  {
    ostrstream os;
    if (width>0) {
      os.fill('0');
      os.width(width);
      os.setf(ios::internal, ios::adjustfield);
    }
    os << n << '\0';
    return os.str();
  }


}

#endif
